
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include "util.h"
 
#define MinEnsembleSize   3
#define MaxEnsembleSize  50
#define ValidateFraction  3
#define EarlyStopMargin   1.5
#define MinIteration    250
#define min(a,b) (((a)<(b))?(a):(b))

void createNet();
void destroyNet();
void randomizeNet(real magnitude);
real prior(real *dw) { return 0.0; }        /* we do not have a weight prior */
extern void initTimer(struct itimerval *timer, int time);
extern long elapsedTime(struct itimerval *timer);
real cost(struct exampleset ex, int wc);
void net(real *inp, real *out);
extern int conj(int *iter, int *epoch, int *restart, real *costvalue);

struct itimerval timer;   
struct exampleset train;                            /* the training examples */

real *w,                            /* array [no_wts] of all network weights */
     **wi,   /* array [no_hid] of arrays [no_inp] of input-to-hidden weights */
     **wd,                        /* array of direct input-to-output weights */
     **wo,                              /* array of hidden-to-output weights */
     *output;                                    /* array of network outputs */
int  no_hid, no_inp, no_tar,        /* no. of hidden, input and target units */
     no_wts;                                         /* total no. of weights */


main(int argc, char **argv)
{
  
  char df[50];                                                  /* file name */
  long seed,             
       length, used_time = 0;                   /* times are in milliseconds */
  int  restart,    /* used with conj(); if 1 then conj uses steepest descent */
       i, j, k, l,                                  
       iter, best_iter,     /* current and best iteration for early stopping */
       ensembleSize,                 /* counts number of nets trained so far */
       succ; 
  real *tmp,
       junk,
       perf, best_perf,   /* current and best performance for early stopping */
       **pred,                                       /* array of predictions */
       *best_w;            /* array of weights of best net in early stopping */
  struct exampleset valid, test; 
  extern int TIMEOUT;  /* set to 1 when SIGVTALRM is caught; when time is up */
  FILE *wgtf,                /* file for storing ensemble of network weights */
       *pred_file;                   /* file to which predictions are dumped */
  
  
  length = -1875;          /* default length is 1.875 secs per training case */
  time(&seed);                                 /* chose a seed based on time */
  no_inp = no_tar = -1;    /* defaults for "unknown" in case of no arguments */
  if (argc < 2 || argc > 4 ||
      argc > 2 && ((length = 1000*atof(argv[2])) == 0.0) ||  /* length in ms */
      argc==4 && ((seed = atoi(argv[3])) <= 0)) {
    fprintf(stderr, "Usage: %s instance-number [[-]cpu-time [seed]]\n", argv[0]);
    exit(-1);
  }
  srand48(seed); 
  
  sprintf(df, "test.%s", argv[1]);                       /* read test inputs */
  test.num = train.num = no_inp = no_tar = -1;      /* default for "unknown" */
  loadExamples(&test, &no_inp, (i=0, &i), df, NULL);
  sprintf(df, "train.%s", argv[1]);
  loadExamples(&train, &no_inp, &no_tar, df, NULL);
  if (length < 0) length *= -train.num;  /* training time was given per case */
  valid.num = train.num/ValidateFraction;
  train.num -= valid.num;         /* now train on only the training fraction */
  valid.inp = &train.inp[train.num]; valid.tar = &train.tar[train.num];
  
  
  sprintf(df, "wgts.%s", argv[1]); 
  if ((wgtf = fopen(df, "w+b")) == NULL) {
    fprintf(stderr, "Could not open weight file %s for writing ...bye!\n", df);
    exit(-1);
  } 
  
  no_hid = ceil((train.num-(no_inp+1.0)*no_tar)/(no_inp+no_tar+1.0));
  no_wts = (no_inp+1)*no_hid+(no_inp+no_hid+1)*no_tar;
  fprintf(stderr, "Using nets with %3d hidden units and %5d weights.\n",
          no_hid, no_wts);
  createNet();
  best_w = (real *) malloc((size_t) no_wts*sizeof(real)); 
  ensembleSize = 0;
  fprintf(stderr, "net  train-err valid-err  best-iter  max-iter\n");  
  while (used_time < length && ensembleSize < MaxEnsembleSize) {
    
    for (i=0; i<train.num+valid.num; i++) {      
      j = drand48()*train.num;
        tmp = train.inp[j]; train.inp[j] = train.inp[i]; train.inp[i] = tmp;
        tmp = train.tar[j]; train.tar[j] = train.tar[i]; train.tar[i] = tmp;
      }
    
    randomizeNet(0.1);
    initTimer(&timer, min(length/MinEnsembleSize, length-used_time)); 
    best_perf = cost(valid, 0);               /* initialize best_perf and... */
    for (i=0; i<no_wts; i++) best_w[i] = w[i];   /* best_w to the random net */
    iter = best_iter = 0; restart = 1;   /* conj() should restart its search */
    do {
      succ = conj((i=1, &i), (j=0, &j), &restart, &junk); iter++;
      if ((perf=cost(valid, 0)) < best_perf) {
        best_iter = iter; best_perf = perf;          /* update "best" values */
        for (i=0; i<no_wts; i++) best_w[i] = w[i];          /* save this net */
      }    
    } while ((succ) && (!TIMEOUT) && ((iter < EarlyStopMargin*best_iter) ||
     (iter < MinIteration) || (MaxEnsembleSize*elapsedTime(&timer) < length)));
    used_time += elapsedTime(&timer);
    fwrite(best_w, no_wts*sizeof(real), 1, wgtf);      /* append net to wgtf */
    if (used_time<length || ensembleSize<MinEnsembleSize) {   /* add if time */
      for (i=0; i<no_wts; i++) w[i] = best_w[i];
      fprintf(stderr, "%3d %9.4f %9.4f %7d    %7d\n", ++ensembleSize,
                                   cost(train, 0), best_perf, best_iter, iter);
    }
  }
  
  rewind(wgtf);                              /* compute ensemble predictions */
  pred = createMatrix(no_tar, test.num);
  for (k=0; k<no_tar; k++) for (i=0; i<test.num; i++) pred[k][i] = 0.0;
  
  for (j=0; j<ensembleSize; j++) {
    fread(w, (long) no_wts*sizeof(real), 1, wgtf); 
    for (i=0; i<test.num; i++) {
      net(test.inp[i], output);
      for (k=0; k<no_tar; k++) pred[k][i] += output[k];
    }
  }
  
  for (k=0; k<no_tar; k++)                            /* use mean prediction */
    for (i=0; i<test.num; i++) pred[k][i] /= ensembleSize;
  fclose(wgtf); 
  
  pred_file = openPredFile("cguess.%s", argv[1]);   /* point prediction file */
  for (i=0; i<test.num; i++) {                          /* write predictions */
    for (k=0; k<no_tar; k++) fprintf(pred_file, "%f ", pred[k][i]);
    fprintf(pred_file, "\n");
  } 
  fclose(pred_file); free(pred[0]); free(pred);
  
} 
